from . import Basic_tuner
from abc import abstractmethod
from typing import Literal

from joblib import parallel_config

from sklearn.svm import SVR
from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.linear_model import ElasticNet
from xgboost import XGBRegressor
from lightgbm import LGBMRegressor, early_stopping

import shap

import numpy as np
from statsmodels.regression.linear_model import OLS

# Todo: Now loss(not metrics) only support mse, we need more like mae, mape, hinge... etc


class Regression_tuner(Basic_tuner):
    """
    A subclass of Basic_tuner for regression task.
    
    """

    def __init__(self,
                 n_try,
                 n_cv,
                 target,
                 validate_penalty,
                 TT_coef,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def is_regression(self) -> bool:
        """
        Yes, it is for regression.

        Returns:
            bool: True
        """
        return True


# linear model, elasticnet
class ElasticNet_tuner(Regression_tuner):
    """
    Tuning a elasic net regression.    
    [sklearn.linear_model.ElasticNet](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html)
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mse",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "ElasticNet"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html"

        return refer

    def parms_range(self) -> dict:
        return {
            'alpha': ('alpha', "float", 1e-6, 1e+2),
            'l1_ratio': ('l1_ratio', "float", -0.5, 1.5)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {"random_state": self.kernel_seed, "selection": "random"}
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                if par == "l1_ratio":
                    parms[par] = np.clip(
                        self.parms_range_sparser(trial, parms_to_tune[par]), 0,
                        1)
                else:
                    parms[par] = self.parms_range_sparser(
                        trial, parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        enr = ElasticNet(**parms)
        return enr

    def summary(self):
        """
        It is the way I found to cram a sklearn regression result into the statsmodel regresion.    
        The only reason to do this is that statsmodel provides R-style summary.    
        """
        raise NotImplementedError
        sm_ols = OLS(self.y, self.x).fit_regularized(
            alpha=self.best_model.alpha,
            L1_wt=self.best_model.l1_ratio,
            start_params=self.best_model.coef_.flatten(),
            maxiter=0,
        )
        print(sm_ols.summary())

    def _explainer(self, x):
        return shap.LinearExplainer(self.best_model, x)


# RF
class RandomForest_tuner(Regression_tuner):
    """
    Tuning a random forest model.    
    [sklearn.ensemble.RandomForestRegressor](https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html)
    """

    def __init__(self,
                 using_oob=True,
                 n_try=50,
                 n_cv=5,
                 target="mse",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        """
        Args:
            using_oob (bool, optional): Using out of bag score as validation. Defaults to True.
        """
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

        self.using_oob = using_oob

    def name(self):
        return "RandomForest"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html"
        refer[
            self.name() +
            " publication"] = "https://www.stat.berkeley.edu/~breiman/randomforest2001.pdf"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 32, 512),
            'min_samples_leaf':
            ('min_samples_leaf', "int", 1, round(np.sqrt(self.n_sample) / 2)),
            'ccp_alpha':
            ('ccp_alpha', "float", 1e-2 / self.n_sample, 1e+2 / self.n_sample),
            'max_samples': ('max_samples', "float", 0.4, 0.8),
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) * 2)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "bootstrap": True,
            "oob_score": self.using_oob,
            "n_jobs": -1,
            "random_state": self.kernel_seed,
            "verbose": 0,
            "max_features": "sqrt",
        }

        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        rf = RandomForestRegressor(**parms)
        return rf

    def evaluate(self, trial, default=False, training=False):
        regressor_obj = self.create_model(trial, default)

        if self.using_oob:
            # oob predict
            with parallel_config(backend='loky'):
                regressor_obj.fit(self.x, self.y)
            y_pred = regressor_obj.oob_prediction_

            # oob score
            score = self.metric._score_func(self.y, y_pred)

            if not self.metric_great_better:
                # if not great is better, then multiply -1
                score *= -1
        else:
            with parallel_config(backend='loky'):
                score = super().evaluate(trial=trial,
                                         default=default,
                                         training=training)
        return score

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# SVM
class SVM_tuner(Regression_tuner):
    """
    Tuning a support vector machine.    
    [sklearn.svm.SVR](https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html)
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mse",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "SVM"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html"
        refer[
            self.name() +
            " sklearn backend"] = "http://www.csie.ntu.edu.tw/~cjlin/papers/libsvm.pdf"
        refer[
            self.name() +
            " publication"] = "https://citeseerx.ist.psu.edu/doc_view/pid/42e5ed832d4310ce4378c44d05570439df28a393"

        return refer

    def parms_range(self) -> dict:
        # scaling penalty: https://scikit-learn.org/stable/auto_examples/svm/plot_svm_scale_c.html#sphx-glr-auto-examples-svm-plot-svm-scale-c-py
        return {
            "kernel": ('kernel', "category", ["linear", "rbf",
                                              "sigmoid"], None),
            'C': ('C', "float", 1e-3 * np.sqrt(self.n_sample),
                  1e+2 * np.sqrt(self.n_sample))
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "kernel": "rbf",
            'cache_size': 2000,
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
        svm = SVR(**parms)
        return svm

    def _explainer(self, x):
        return shap.KernelExplainer(self.best_model.predict, x)


# XGboost
class XGBoost_tuner(Regression_tuner):
    """
    Tuning a XGBoost classifier model.    
    [xgboost.XGBClassifier](https://xgboost.readthedocs.io/en/stable/python/python_api.html)

    ToDo:    
        1. sample imbalance. (Done)
        2. early stop. (give up)
        3. efficiency (optuna.integration.XGBoostPruningCallback).

    """

    def __init__(self,
                 n_try=50,
                 n_cv=5,
                 target="mse",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "XGBoost"

    def reference(self) -> dict[str, str]:
        """
        This function will return reference of this method in python dict.    
        If you want to access it in PineBioML api document, then click on the    >Expand source code     

        Returns:
            dict[str, str]: a dict of reference.
        """
        refer = super().reference()
        refer[self.name() +
              " document"] = "https://xgboost.readthedocs.io/en/stable/"
        refer[
            self.name() +
            " publication"] = "https://dl.acm.org/doi/10.1145/2939672.2939785"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 4, 256),
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) * 2),
            "gamma": ('gamma', "float", 1e-4, 1e-2),
            "min_child_weight": ("min_child_weight", "float", 1,
                                 round(np.sqrt(self.n_sample) / 2)),
            "learning_rate": ('learning_rate', "float", 1e-2, 1.),
            "subsample": ('subsample', "float", 0.5, 1.),
            "colsample_bytree": ('colsample_bytree', "float", 0.5, 1),
            "reg_lambda": ('reg_lambda', "float", 1e-4, 1),
            "reg_alpha": ('reg_alpha', "float", 1e-4, 1.)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "n_jobs": None,
            "random_state": self.kernel_seed,
            "verbosity": 0
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]

            if training:
                # Adding early stopping callbacks
                parms["early_stopping_rounds"] = round(
                    parms["n_estimators"] * 0.1) + 2
            else:
                # not training, then overide the n_estimators by trial's early stopping point.
                parms["n_estimators"] = self.stop_points[trial.number]

        xgb = XGBRegressor(**parms)
        return xgb

    def using_earlystopping(self):
        return True

    def clr_best_iteration(self, classifier):
        return classifier.best_iteration

    def optimize_fit(self, clr, train_data, sample_weight, valid_data):
        train_x, train_y = train_data

        return clr.fit(train_x,
                       train_y,
                       sample_weight=sample_weight,
                       eval_set=[valid_data],
                       verbose=False)

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# lightGBM
class LightGBM_tuner(Regression_tuner):
    """
    Tuning a LighGBM classifier model.    
    [lightgbm.LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html)     

    ToDo:    
        1. compare with optuna.integration.lightgbm.LightGBMTuner    
    """

    def __init__(self,
                 n_try=50,
                 n_cv=5,
                 target="mse",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "LightGBM"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://lightgbm.readthedocs.io/en/latest/index.html"
        refer[
            self.name() +
            " publication"] = "https://proceedings.neurips.cc/paper_files/paper/2017/file/6449f44a102fde848669bdd9eb6b76fa-Paper.pdf"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 4, 256),
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) * 2),
            "min_child_samples":
            ("min_child_samples", "int", 1, int(np.sqrt(self.n_sample) / 2)),
            "learning_rate": ('learning_rate', "float", 1e-2, 1.),
            "subsample": ('subsample', "float", 0.5, 1.),
            "colsample_bytree": ('colsample_bytree', "float", 0.5, 1.),
            "reg_lambda": ('reg_lambda', "float", 1e-4, 1),
            "reg_alpha": ('reg_alpha', "float", 1e-4, 1.)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {
            "n_jobs": None,
            "random_state": self.kernel_seed,
            "verbosity": -1,
            "subsample_freq": 1
        }
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
            if not training:
                # not training, then overide the n_estimators by trial's early stopping point.
                parms["n_estimators"] = self.stop_points[trial.number]

        lgbm = LGBMRegressor(**parms)
        return lgbm

    def using_earlystopping(self):
        return True

    def clr_best_iteration(self, classifier):
        return classifier.best_iteration_

    def optimize_fit(self, clr, train_data, sample_weight, valid_data):
        train_x, train_y = train_data

        return clr.fit(train_x,
                       train_y,
                       sample_weight=sample_weight,
                       eval_set=[valid_data],
                       callbacks=[
                           early_stopping(round(clr.n_estimators * 0.1) + 2,
                                          verbose=False)
                       ])

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# AdaBoost
class AdaBoost_tuner(Regression_tuner):
    """
    Tuning a AdaBoost regressor.    
    [sklearn.ensemble.AdaBoostRegressor](https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostRegressor.html)
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mse",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        """
        Args:
            n_try (int, optional): The number of trials optuna should try. Defaults to 25.
            n_cv (int, optional): The number of folds to execute cross validation evaluation in iteration of optuna optimization. Defaults to 5.
            target (str, optional): The target of optuna optimization. Notice that is different from the training loss of model. Defaults to "mse".
            kernel_seed (int, optional): Random seed for model. Defaults to None.
            valid_seed (int, optional): Random seed for cross validation. Defaults to None.
            optuna_seed (int, optional): Random seed for optuna's hyperparameter sampling. Defaults to None.
            validate_penalty (bool, optional): True to penalty the overfitting by difference between training score and cv score. Defaults to True.
        """
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "AdaBoost"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostRegressor.html"
        refer[
            self.name() +
            " publication: original version"] = "https://www.ee.columbia.edu/~sfchang/course/svia-F03/papers/freund95decisiontheoretic-adaboost.pdf"
        refer[
            self.name() +
            " publication: implemented version"] = "Drucker, 'Improving Regressors using Boosting Techniques', 1997"

        return refer

    def parms_range(self) -> dict:
        return {
            "n_estimators": ('n_estimators', "int", 4, 64),
            "learning_rate": ('learning_rate', "float", 1e-2, 1.),
            "loss": ("loss", "category", ["linear", "square",
                                          "exponential"], None)
        }

    def create_model(self, trial, default=False, training=False):
        parms = {"random_state": self.kernel_seed}
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        ada = AdaBoostRegressor(**parms)
        return ada

    def _explainer(self, x):
        return shap.KernelExplainer(self.best_model.predict, x)


# DT
class DecisionTree_tuner(Regression_tuner):
    """
    Tuning a DecisionTree regressor.    
    [sklearn.tree.DecisionTreeRegressor](https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html)
    """

    def __init__(self,
                 n_try=25,
                 n_cv=5,
                 target="mse",
                 validate_penalty=True,
                 TT_coef=0.1,
                 kernel_seed=None,
                 valid_seed=None,
                 optuna_seed=None):
        super().__init__(n_try=n_try,
                         n_cv=n_cv,
                         target=target,
                         validate_penalty=validate_penalty,
                         TT_coef=TT_coef,
                         kernel_seed=kernel_seed,
                         valid_seed=valid_seed,
                         optuna_seed=optuna_seed)

    def name(self):
        return "DecisionTree"

    def reference(self) -> dict[str, str]:
        refer = super().reference()
        refer[
            self.name() +
            " document"] = "https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html"
        refer[
            self.name() +
            " publication"] = "https://www.taylorfrancis.com/books/mono/10.1201/9781315139470/classification-regression-trees-leo-breiman-jerome-friedman-olshen-charles-stone"

        return refer

    def parms_range(self) -> dict:
        return {
            "max_depth":
            ("max_depth", "int", round(np.log2(self.n_sample) / 2),
             int(np.log2(self.n_sample)) + 2),
            "min_samples_leaf":
            ('min_samples_leaf', "int", 1, round(np.sqrt(self.n_sample) / 2)),
            'ccp_alpha':
            ('ccp_alpha', "float", 1e-2 / self.n_sample, 1e+2 / self.n_sample),
        }

    def create_model(self, trial, default=False, training=False):
        parms = {"random_state": self.kernel_seed}
        if not default:
            parms_to_tune = self.parms_range()
            for par in parms_to_tune:
                parms[par] = self.parms_range_sparser(trial,
                                                      parms_to_tune[par])
            parms["random_state"] = self.kernel_seed_tape[trial.number]
        DT = DecisionTreeRegressor(**parms)
        return DT

    def _explainer(self, x):
        return shap.TreeExplainer(self.best_model)


# CatBoost
# KNN



from typing import Union
from pandas import read_csv, read_excel, read_table, concat, Series, DataFrame
from numpy import ones
import joblib
import os


def read_file(
    file_path: str,
    index_col: Union[int,
                     str] = None) -> Union[DataFrame, dict[str, DataFrame]]:
    """
    Read files, which supports common data format .csv, .tsv, .xlsx and R-table in .txt format.    
    Notice that while an excel file has several data sheet, this function will return a python dict with corresponding sheet name as dict keys.    

    Args:
        file_path (str): The path of target file.
        index_col (Union[int, str], optional): The column index or name to set as the index of dataframe. Set None to ignore. Defaults to None.

    Returns:
        Union[DataFrame, dict[str, DataFrame]]: While the target file is an excel with more than one data sheet, it will return a dict consisting of the data sheets.
    """

    # sparse csv, tsv or excel
    file_type = file_path.split(".")[-1]
    if file_type == "csv":
        file = read_csv(file_path, index_col=index_col, engine="pyarrow")
    elif file_type == "tsv":
        file_a = read_csv(file_path,
                          sep=" ",
                          index_col=index_col,
                          engine="pyarrow")
        file_b = read_csv(file_path,
                          sep="\t",
                          index_col=index_col,
                          engine="pyarrow")
        if file_a.isna().mean().mean() > file_b.isna().mean().mean():
            file = file_b
        else:
            file = file_a

    elif file_type in ["xls", "xlsx", "xlsm", "xlsb"]:
        file = read_excel(file_path, sheet_name=None, index_col=index_col)
        if len(file) == 1:
            file = file[list(file.keys())[0]]

    elif file_type == "txt":
        file = read_table(file_path, index_col=index_col, engine="pyarrow")
    else:
        raise NotImplementedError(
            "The type of target file is not supported. Must be one of .csv, .tsv, .xls, .xlsx or R table in .txt format."
        )
    return file


def read_multiple_groups(
        file_path_list: list[str],
        transpose: bool = False,
        index_col: Union[int, str] = None) -> tuple[DataFrame, Series]:
    """

    read_multiple_groups will do:    
        1. read files in    
        2. assign y according to the read-in order    
        3. concatenate all datas in row    
    It sould be used while the data was sperated into different files by their class or some kind of property we interested in.    

    Args:
        file_path_list (list[str]): A list of path to target files.
        transpose (bool, optional): Transpose before concatenating. Defaults to False.
        index_col (Union[int, str], optional): The column index or name to set as the index of dataframe. Set None to ignore. Defaults to None.

    Returns:
        tuple[DataFrame, Series]: A tuple (x, y) where x is the stacking of target files, and y is an array of integer which records each row's source in the order of file_path_list.    
    """

    # 1. read files in
    datas = []
    group_label = []
    label = 0
    for path in file_path_list:
        group = read_file(path, index_col)
        if transpose:
            group = group.T

        datas.append(group)
        # 2. assign y according to the read-in order
        group_label.append(
            Series(ones(group.shape[0]) * label, index=group.index))
        label += 1

    # 3. concatenate all the datas in row
    x = concat(datas, axis=0)
    y = concat(group_label, axis=0)

    if not len(set(y.index)) == len(y.index):
        # if index repeats, drop them all.
        x = x.reset_index(drop=True)
        y = y.reset_index(drop=True)
    
    y.name = 'label'

    return x, y


def save_model(model_obj: object,
               save_path: str,
               save_name: str,
               overide: bool = False) -> None:
    """
    Saving the model_obj by joblib in pickle format to the path: save_path/save_name .    
    If overide, then overide while any file with the same save_name has already existed in the save_path.

    Args:
        model_obj (object): The model or pipeline to be saved.
        save_path (str): The path to to save the model_obj.
        save_name (str): the saving filename.
        overide (bool, optional): True for overidind any file with save_name in save_path. Defaults to False.
    """
    if not save_path[-1] == "/":
        save_path = save_path + "/"

    if not os.path.exists(save_path):
        print(save_path, " does not exist yet. we will try to create it.")
        os.makedirs(save_path)

    if os.path.exists(save_path + save_name):
        print(save_name, " has already exist in ", save_path)
        if overide:
            print("It will be overide.")
            joblib.dump(model_obj, save_path + save_name)
        else:
            print(
                "please choose another model save_name or set overide to True which will replace the existing one"
            )
    else:
        joblib.dump(model_obj, save_path + save_name)


def load_model(save_path: str, save_name: str = None) -> object:
    """
    Load a saved model.    

    If save_name was provided, load_model will try to access save_path/save_name .    
    else save_path will be regarded as save_path/save_name (lazy mode).    

    For example, a model saved as "./output/model/best_model"    
        load_model("./output/model/", "best_model")    
        load_model("./output/model/best_model")    
    have exact the same result.

    Args:
        save_path (str): The path to folder where the model saved.
        save_name (str, optional): The name of target model. Defaults to None.

    Returns:
        object: An python object that was saved in previous. DO NOT TRUST UNKNOWN SOURCE PICKLE FILES.
    """
    if not save_name is None:
        if not save_path[-1] == "/":
            save_path = save_path + "/"

        return joblib.load(save_path + save_name)
    else:
        # lazy mode
        return joblib.load(save_path)



from sklearn.decomposition import PCA

MAX_ITEM = 32
BAGGING_THRESHOLD = 0.9


class sparser_node:
    """ 
    The kernel of string tree dictionary. It will sparse the feature name and then bagging similar feature with high corelation.
    
    For all feature names of a node in k-th layer, their first k digits will be the same.

    For all feature names of a node, they will be bagged together.

    """

    def __init__(self, words, k):
        """
        Args:
            words (_type_): strings to sparse
            k (_type_): deepth of node.
        """
        self.k = k  # the depth of tree, also the digit of current sparser
        self.name = words[0][:self.k] + "*"  # name of this node
        self.bagged = False  # if the node is a bag, determined after call bagging function

        items = []
        collector = {}
        # separate the words which is not long enough
        # for example, got words = ["a", "b", "a1", "a2"]
        # then the items of this bag should be ["a", "b"]
        # the collector of child will be ["a1", "a2"]
        for word in words:
            if len(word) > self.k + 1:
                if word[self.k] not in collector:
                    collector[word[self.k]] = []
                collector[word[self.k]].append(word)
            else:
                items.append(word)

        self.items = items
        self.children = {}
        # ["a1", "a2"] => {"a": [a1, a2]}
        for symbol in collector:
            self.children[symbol] = sparser_node(collector[symbol], self.k + 1)

    def __call__(self, idx):
        """
        query the child.

        Args:
            idx (char): char to query

        Returns:
            sparse_node: Child coresponding to idx
        """
        return self.children[idx]

    def bagging(self, df):
        """
        Compute the explained variance ratio of 1st PC in PCA of all coresponding features in this node and its child.
        If the ratio > 0.9, the this node will be bagged. 
        The node name will be bagged feature name and 1st pc will be corespoding feature value.

        Args:
            df (_type_): The dataframe of all data.

        Returns:
            list: the list of all items in this node
        """
        items = []
        items += self.items

        for child in self.children:
            items += self.children[child].bagging(
                df)  # recurrently call bagging (deep first, digit tree)

        n = len(items)  # only 1 item => no bagging
        if n == 1:
            return items
        elif n < MAX_ITEM:  # max number of items for each bag
            data = df[items]
            ### bagging rules
            # rule 1
            #score = np.mean(data.corr().abs())
            # rule 2
            pca = PCA(n)
            tmp_data = pca.fit_transform((data - data.mean() / data.std()))
            score = pca.explained_variance_ratio_[0]

            if score > BAGGING_THRESHOLD:
                #print(np.mean(data.corr().abs()), pca.explained_variance_ratio_, "#################")
                #df[self.name] = df[items].mean(axis = 1) # act 1
                df[self.name] = tmp_data[:, 0]  # act 1

                self.bagged = True  # record this node is bagged
                return [self.name]  # bag name
        return items

    def collect(self, return_bag=True):
        """
        collect all items in its children and itself.

        Args:
            return_bag (bool, optional): True will return in bagged result, otherwise not. Defaults to True.

        Returns:
            list: the list of all items in this node and its children.
        """
        if self.bagged and return_bag:
            return [self.name]

        items = []
        items += self.items

        for child in self.children:
            items += self.children[child].collect(return_bag)
        return items

    def see(self, idx):
        """
        Query a specific node.
        
        if idx is empty, then the target is found and its item will be return
        else query its child by the first digit and pop the digit.

        Args:
            idx (string): node name

        Returns:
            list or node: if idx is empty, then the target is found and its item will be return. else query its child by the first digit and pop the digit.
            
        """
        if len(idx) > 1:
            return self(idx[0]).see(idx[1:])
        if idx == "*":
            return self.collect(return_bag=False)
        else:
            for i in self.items:
                if i[-1] == idx:
                    return [i]


class bagger():
    """
    A wrapper of sparser_node
    """

    def __init__(self):
        self.root = None
        self.fitted = False

    def fit(self, x, y=None):
        """
        Fit the string tree dictionary.
        Args:
            x (pandas.DataFrame): dataframe. should have .columns attribute
            y (pandas.Serise or 1D array, optional): label or target. This have no effects. Defaults to None.

        Returns:
            bagger: Fitted self.
        
        To do:
            transform
        """
        # sparse the column names
        self.root = sparser_node(x.columns, 0)

        # bagging
        self.template_x = x.copy()
        self.bags = self.root.bagging(self.template_x)

        self.fitted = True
        return self

    """
    def transform(self, x, y=None):
        
        if not self.fitted:
            raise "please call fit before calling transform."

        return self.template_x[self.bags], y
    """

    def fit_transform(self, x, y=None):
        """
        A functional stack of fit and calling the result

        Args:
            x (_type_): _description_
            y (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        self.fit(x, y)
        return self.template_x[self.bags], y  #self.transform(x, y)

    def see(self, idx):
        """
        A functional interface of sparser_node.see

        Args:
            idx (string): see sparser_node.see

        Returns:
            node or list: see sparser_node.see
        """
        return self.root.see(idx)

    def unbagging(self, name):
        """
        inverse transform of bagging

        Args:
            name (string): bagging name to query

        Returns:
            list: items in this bag.
        """
        return " ".join(self.see(name))



from sklearn import metrics
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.pipeline import Pipeline
from pandas import DataFrame, Series, concat
from sklearn.base import is_classifier, is_regressor
import time
from sklearn.base import BaseEstimator


class data_source():
    """
    The data_source is a placeholder for various data source. it will override the data flow of Pine.
    Notice that the label y and the order should the same for each datasource.
    """

    def __init__(
        self,
        train_x: DataFrame,
        test_x: DataFrame = None,
    ):
        self.train_x = train_x
        self.test_x = test_x

    def fit(self, x: DataFrame, y: Series):

        return self

    def transform(self, x: DataFrame):

        return self.test_x

    def fit_transform(self, x: DataFrame, y: Series):

        return self.train_x


class sklearn_esitimator_wrapper(BaseEstimator):
    """
    A basic wrapper for sklearn_esitimator. It transfer the data pipeline of sklearn from numpy.array to pandas.DataFrame.    
    If you want to pass any model with api in sklearn style into Pine, you should wrap it in wrapper.
    """

    def __init__(self, kernel: object):
        """

        Args:
            kernel (object): a sklearn esitimator. for example: sklearn.ensemble.RandomForestClassifier or sklearn.ensemble.RandomForestRegressor
        """

        self.kernel = kernel

    def fit(self, x: DataFrame, y: Series, retune=True) -> object:
        """
        sklearn esitimator api: fit

        Args:
            x (DataFrame): feature
            y (Series): label
            retune (bool, optional): To retune the model or not. For sklearn_esitimator_wrapper, it is just a placeholder without acutual facility. Defaults to True.

        Returns:
            object: A sklearn_esitimator within pandas data flow.
        """
        self.label_name_ = y.name
        self.kernel.fit(x, y)
        return self

    def predict(self, x: DataFrame) -> Series:
        """
        sklearn esitimator api: predict

        Args:
            x (DataFrame): feature

        Returns:
            Series: kernel prediction
        """

        return Series(self.kernel.predict(x),
                      index=x.index,
                      name=self.label_name_)

    def predict_proba(self, x: DataFrame) -> DataFrame:
        """
        sklearn esitimator api: predict_proba for classification

        Args:
            x (DataFrame): feature

        Raises:
            NotImplementedError: Regression has no attribute 'predict_proba'

        Returns:
            DataFrame: predicted probability with shape (n_sample, n_class)
        """
        if "predict_proba" in dir(self.kernel):
            return DataFrame(self.kernel.predict_proba(x),
                             index=x.index,
                             columns=self.kernel.classes_)
        else:
            raise NotImplementedError(
                "{} do not have attribute 'predict_proba'.".format(
                    self.kernel.__str__()))

    def is_regression(self) -> bool:
        return is_regressor(self.kernel)

    def detail(self):
        return None


class classification_scorer():
    """
    A utility to calculate classification scores.    
    The result will contain mcc(matthews corrcoef), acc(accuracy) and support(the number of samples), furthermore:    
        if target_label was given(not None), then sensitivity, specificity and coresponding roc-auc score will be added to result.    
        if multi_class_extra is True, then one vs rest macro_auc, cross_entropy and cohen_kappa will be added to result.    
    """

    def __init__(self,
                 target_label: str = None,
                 prefix: str = "",
                 multi_class_extra: bool = False):
        """

        Args:
            target_label (str, optional): the name of target_label. For example, the label in a binary classification task might be {'pos', 'neg'}. Then you can assign 'neg' to target_label, and the result will contain sensitivity, specificity and roc-auc score of label 'neg'. Defaults to None.
            prefix (str, optional): prefix before score names. For example suppose prefix="Train_", then all the scores in result will be like "Train_mcc". Defaults to "".
            multi_class_extra (bool, optional): _description_. Defaults to False.
        """

        self.prefix = prefix
        self.target_label = target_label
        self.multi_class_extra = multi_class_extra

    def score(self,
              y_true: Series,
              y_pred_prob: DataFrame,
              y_pred: Series = None) -> dict[str, float]:
        """
        Scoring y_true and y_pred_prob.

        Args:
            y_true (Series): The ground True.
            y_pred_prob (DataFrame): The prediction from an estimator. Shape should be (n_sample, n_classes)
            y_pred (Series, optional): The prediction made by model. For Binary classification models, the prediction may differ from prob.argmax because of threshold tuning. Defaults to None.

        Returns:
            dict[str, float]: The result stored in a dict, be like {'score_name': score}.
        """
        if y_pred is None:
            y_pred = y_pred_prob.idxmax(axis=1)

        result = {}
        if not self.target_label is None:
            (_, result["sensitivity"], result["f1"],
             _) = metrics.precision_recall_fscore_support(
                 y_true=(y_true == self.target_label),
                 y_pred=(y_pred == self.target_label),
                 average="binary",
                 pos_label=True)

            (_, result["specificity"], _,
             _) = metrics.precision_recall_fscore_support(
                 y_true=(y_true == self.target_label),
                 y_pred=(y_pred == self.target_label),
                 average="binary",
                 pos_label=False)

            # binary
            result["auc"] = metrics.roc_auc_score(
                y_true == self.target_label, y_pred_prob[self.target_label])

        if self.multi_class_extra:
            result["macro_auc"] = metrics.roc_auc_score(
                y_true,
                y_pred_prob,
                multi_class="ovr",
                labels=y_pred_prob.columns)
            result["macro_f1"] = metrics.f1_score(y_true,
                                                  y_pred,
                                                  average="macro")
            result["cross_entropy"] = metrics.log_loss(y_true, y_pred_prob)
            result["cohen_kappa"] = metrics.cohen_kappa_score(y_true, y_pred)

        result["mcc"] = metrics.matthews_corrcoef(y_true, y_pred)
        result["accuracy"] = metrics.accuracy_score(y_true, y_pred)
        result["support"] = len(y_true)

        prefix_result = {}
        for score in result:
            prefix_result[self.prefix + score] = result[score]

        return prefix_result


class regression_scorer():
    """
    A utility to calculate regression scores. rmse(rooted mean squared error), r2(R squared) and support(the number of samples) are included.    
    if y_true and y_pred are all positive, then mape(mean absolute percentage error) will be added.    
    """

    def __init__(self, prefix: str = "", target_label: str = None):
        """

        Args:
            prefix (str, optional): prefix before score names. For example suppose prefix="Train_", then all the scores in result will be like "Train_mse". Defaults to "".
            target_label (str, optional): A placehold without any facility. Defaults to None.
        """

        self.prefix = prefix

    def score(self,
              y_true: Series,
              y_pred: Series,
              place_holder=None) -> dict[str, float]:
        """
        calculate the scores

        Args:
            y_true (Series): Ground true.
            y_pred (Series): predicted values.
            place_holder (None): A placeholder corresponding to classification_scorer's pred argument.

        Returns:
            dict[str, float]: The result stored in a dict, be like {'score_name': score}.
        """

        result = {}
        result["rmse"] = metrics.root_mean_squared_error(y_true, y_pred)
        result["r2"] = metrics.r2_score(y_true, y_pred)
        if (y_true > 0).all() and (y_pred > 0).all():
            result["mape"] = metrics.mean_absolute_percentage_error(
                y_true, y_pred)
        result["support"] = len(y_true)

        prefix_result = {}
        for score in result:
            prefix_result[self.prefix + score] = result[score]

        return prefix_result


# ToDo: the Integration of .predict and .transform
class Pine():
    """
    Deep first traversal the given experiment setting.    
    the last step of experiment sould be model.    
    Please refer to example_Pine.ipynb for usage.    


    note: experiment step and experiment stage is the same thing.    
    """

    def __init__(self,
                 experiment: list[tuple[str, dict[str, object]]],
                 target_label: str = None,
                 cv_result: bool = False,
                 evaluate_ncv: int = 5):
        """
        Args:
            experiment (list[tuple[str, dict[str, object]]]): list of experiment steps. step should be in the form: ('step_name', {'method_name': method}). it could be several method in one step and they will fork in deep first traversal. Each method should be either sklearn estimator or transformer.
            target_label (str, optional): the name of target_label. For example, the label in a binary classification task might be {'pos', 'neg'}. Then you can assign 'neg' to target_label, and the result will contain sensitivity, specificity and roc-auc score of label 'neg'. Defaults to None.
            cv_result (bool, optional): Rcording the scores and prediction of cross validation. Defaults to False.
            evaluate_cv (int, optional): The number of folds to evaluate cv_result after pipeline tuned. Defaults to 5.
        """

        self.experiment = experiment
        self.total_stage = len(experiment)
        self.target_label = target_label
        self.cv_result = cv_result
        self.evaluate_ncv = evaluate_ncv

        self.result = []

        self.train_pred = []
        self.cv_pred = []
        self.test_pred = []

    def do_stage(self, train_x: DataFrame, train_y: Series, test_x: DataFrame,
                 test_y: Series, stage: int, record_path: dict,
                 record_time: dict) -> None:
        """
        the recursive function to traversal the experiment.    
        the socres and path will be stored in self.result amd self.____pred, so there is no return in recursive function.     

        Args:
            train_x (pd.DataFrame): training x
            train_y (pd.Series): training y
            test_x (pd.DataFrame): training x
            test_y (pd.Series): training y
            stage (int): the order of current stage in the experiment setting
            record_path (dict): record_path the traversal path in a dict of str
            record_time (dict): record_time the traversal time in a dict of str
        """

        # unzip the stage, stage = (stage_name, {operator_name: operator})
        stage_name, operators = self.experiment[stage]

        # fork to next stage according to the diffirent operator (opt)
        for opt_name in operators:
            record_path[stage_name] = opt_name

            opt = operators[opt_name]

            # if not the last stage
            if stage < self.total_stage - 1:
                time_start = time.time()

                # transform by operators
                processed_train_x = opt.fit_transform(train_x, train_y)
                if test_x is not None:
                    processed_test_x = opt.transform(test_x)
                else:
                    processed_test_x = test_x

                time_end = time.time()
                record_time[stage_name + "_time"] = time_end - time_start
                # reccursivly call
                self.do_stage(processed_train_x, train_y, processed_test_x,
                              test_y, stage + 1, record_path, record_time)

            # the last layer, it should be models
            else:
                model = opt
                if "predict_proba" in dir(model):
                    # is not regression
                    f = model.predict_proba
                    scorer = classification_scorer
                else:
                    # is regression
                    f = model.predict
                    scorer = regression_scorer
                time_start = time.time()
                # tune/fit the model on training data
                model.fit(train_x, train_y)
                time_end = time.time()
                record_time[stage_name + "_fit_time"] = time_end - time_start

                # compute the training score
                time_start = time.time()
                train_pred = f(train_x)
                time_end = time.time()
                record_time[stage_name +
                            "_predict_time"] = time_end - time_start

                # compute the prediction for those who has a tuned threshold in binary classification task.
                train_prediction = model.predict(train_x)

                self.train_pred.append(train_pred)
                train_scores = scorer(prefix="train_",
                                      target_label=self.target_label).score(
                                          train_y, train_pred,
                                          train_prediction)

                if test_x is not None:
                    # if there is testing data, compute the testing score.
                    test_pred = f(test_x)
                    test_prediction = model.predict(test_x)
                    self.test_pred.append(test_pred)
                    test_scores = scorer(prefix="test_",
                                         target_label=self.target_label).score(
                                             test_y, test_pred,
                                             test_prediction)
                else:
                    test_scores = {}

                if self.cv_result:
                    # compute the cross validation score on training set
                    fold_scores = []
                    cv_pred = []

                    if model.is_regression():
                        cross_validation = KFold(n_splits=self.evaluate_ncv,
                                                 shuffle=True,
                                                 random_state=133)
                    else:
                        cross_validation = StratifiedKFold(
                            n_splits=self.evaluate_ncv,
                            shuffle=True,
                            random_state=133)

                    for (train_idx, valid_idx) in cross_validation.split(
                            train_x, train_y):

                        # fit on training fold
                        model.fit(train_x.iloc[train_idx],
                                  train_y.iloc[train_idx],
                                  retune=False)

                        # score on testing fold
                        fold_pred = f(train_x.iloc[valid_idx])
                        #fold_prediction = model.predict(train_x.iloc[valid_idx])

                        cv_pred.append(fold_pred)
                        fold_scores.append(
                            scorer(prefix="cv_",
                                   target_label=self.target_label).score(
                                       train_y.iloc[valid_idx], fold_pred))
                    # average the fold scores
                    self.cv_pred.append(concat(cv_pred, axis=0))
                    valid_scores = DataFrame(fold_scores).mean().to_dict()
                    # TODO accurate statistic estimate of std.
                    valid_std = DataFrame(fold_scores).std().to_dict()
                    valid_std = {f"{k}_std": v for k, v in valid_std.items()}
                else:
                    valid_scores = {}
                    valid_std = {}

                # re-fit
                model.fit(train_x, train_y, retune=False)

                # concatenate the score dicts
                all_scores = dict(**record_path, **record_time, **train_scores,
                                  **valid_scores, **valid_std, **test_scores)
                self.result.append(all_scores)

    def do_experiment(self, train_x, train_y, test_x=None, test_y=None):
        """
        the first call of recurssive fuction.

        Args:
            train_x (pd.DataFrame): training x
            train_y (pd.Series): training y
            test_x (pd.DataFrame): training x
            test_y (pd.Series): training y

        Returns:
            pd.DataFrame: the result
        """
        # clear the results.
        self.result = []

        self.do_stage(train_x, train_y, test_x, test_y, 0, {}, {})
        return self.experiment_results()

    def experiment_results(self, timer=False, std=False) -> DataFrame:
        """
        Args:
            timer (bool): To return the time records.
            std (bool): To return the cv std.

        Returns:
            DataFrame: The experiment results.
        """
        result = DataFrame(self.result)
        to_drop = []
        if not timer:
            to_drop += [i for i in result.columns if i[-5:] == "_time"]
        if not std:
            to_drop += [i for i in result.columns if i[-4:] == "_std"]

        if len(to_drop) == 0:
            return result
        else:
            return result.drop(to_drop, axis=1)

    def experiment_predictions(self):
        """
        cv_pred will be empty if cv_result was False in initialization.

        Returns:
            train_pred, cv_pred, test_pred: the prediction of training set, cross validation and  testing set
        """
        return self.train_pred, self.cv_pred, self.test_pred

    def recall_model(self, id):
        """
        query the last experiment result by id and build the pipeline object.

        Todo: A proper way to fit the pipeline object.

        Args:
            id (int): the order of experiment path.

        Returns:
            sklearn.pipeline.Pipeline: ready to use object.
        """

        model_spec = self.result[id]
        model_pipeline = []
        for (step_name, methods) in self.experiment:
            using_method = model_spec[step_name]
            model_pipeline.append((step_name, methods[using_method]))
        return Pipeline(model_pipeline)

    def experiment_detail(self):
        """
        show the experiment settings including:    
            1. models parameters searching range and results of the last experiment round.

        Returns:
            pandas.DataFrame
        """
        _, models = self.experiment[-1]

        params = []
        for n in list(models):
            m = models[n]
            tmp = m.detail()
            if tmp is not None:
                params.append(tmp)
        return concat(params, axis=0)
